#version 150

in vec3 vertex;
in vec3 normal;
in vec2 uvcord;

uniform int  mode;
uniform mat4 mvpmatrix;
uniform mat4 cammatrix;

out vec2 texcoord;
out float lum;

void main(void){
   const vec3 LIGHTPOS = vec3(0.454, 0.526, 1.155);
   const vec3 MICPOS = vec3(-0.607, 1.642, -0.503);
   const vec3 RATPOS = vec3(-0.1799, 2.4835, -0.6337);

   if( (mode & 1) == 1){
      // Bug perspective
      gl_Position = mvpmatrix * cammatrix * vec4(vertex.xy, vertex.z * 1.3 + 0.4, 1.0);
   }
   else{
      gl_Position = mvpmatrix * cammatrix * vec4(vertex, 1.0);
   }

   if( mode == 2 ){
      /* Green light */
      vec3 L1 = MICPOS - vertex;
      vec3 L2 = RATPOS - vertex;
      float ratdist = max(0.0, 1.0 - distance(vertex, RATPOS)) * 2.3;
      float micdist = max(0.0, 1.0 - distance(vertex, MICPOS)) * 1.6;
      float lum1 = max(0.0, dot(normalize(normal), normalize(L1))) * micdist;
      float lum2 = max(0.0, dot(normalize(normal), normalize(L2))) * ratdist;
      lum = (lum1 + lum2) / 2.0;
   }
   else{
      vec3 lightvector = LIGHTPOS - vertex;
      lum = max(0.0, dot(normalize(normal), normalize(lightvector)));
   }

   texcoord = uvcord;
}
