#define GL_GLEXT_PROTOTYPES
#include <GL/gl.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "biopunk.h"
#include "mesh3d.h"
#include "texture.h"
#include "cg_time.h"
#include "shader.h"
#include "matrix.h"
#include "pdb.h"
#include "celldivide.h"
#include "bone.h"

#define LINESIZE       200
#define PHAGENUM       80
#define PHAGECUBESIZE  9
#define LETTERS        26
#define NORMALMODE     0
#define BUGPERSPECTIVE 1
#define GREENLIGHT     2
#define NORMALRAT      0
#define GREENRAT       1

/* Meshes */
Mesh3D *kitchen;
Mesh3D *walls;
Mesh3D *rat;
Mesh3D *dnasegment;
Mesh3D *tube;
Mesh3D *phage;
Mesh3D *ecoli;
Mesh3D *letters[LETTERS];
PDB     dna;
Mesh3D *logo1;
Mesh3D *logo2;
GLuint  square;
GLuint  squarevbo;

/* Shaders */
GLuint kitchen_src[2];
GLuint kitchen_shader;
GLuint rat_src[2];
GLuint rat_shader; /* Kocka shader (no texture)*/
GLuint dnatwist_src[2];
GLuint dnatwist_shader; /* Shader for DNA Twister */
GLuint letter_src[2];
GLuint letter_shader; /* Shader for letters */
GLuint square_shader; /* Shader for static messages and backgrounds */
GLuint square_src[2];
GLuint twist_shader;/* Shader for DNA twister */
GLuint twist_src[2];
GLuint dof_shader; /* Depth of field shader */
GLuint dof_src[2];
GLuint cell_shader; /* Shader for dividing cells */
GLuint cell_src[2];
GLuint gol_shader; /* Game of life shader */
GLuint gol_src[2];
GLuint simple_shader; /* Show only one texture */
GLuint simple_src[2];

/* Textures */
GLuint kitchentex;
GLuint walltex;
GLuint rattex;
GLuint still;
GLuint rendertexture;
GLuint simplemessage;
GLuint checkmessage;
GLuint goltex;

/* Common matrices */
float perspectivematrix[16];

/* Others */
GLfloat *sniff;
GLuint sniffvbo;
float coliscene[127 * 6];/* Position and rotation of the E. colis */
float phagescene[PHAGENUM * 6];/* Start and stop positions of the phages */
float phagecube[PHAGECUBESIZE][PHAGECUBESIZE][PHAGECUBESIZE][3];
GLuint framebuffer;
GLuint depthbuffer;
int w;
int h;

Bone *bone;

void randomSpeed(){
   int i,j,k,l;

   for(i = 0; i < PHAGECUBESIZE; i++)
      for(j = 0; j < PHAGECUBESIZE; j++)
         for(k = 0; k < PHAGECUBESIZE; k++)
            for(l = 0; l < 3; l++){
               phagecube[i][j][k][l] = drand48() - 0.5;
            }
}

void generateFractal(int x, int y, int z, int dim){
   int i,j,k;

   if(dim == 1){
      phagecube[x][y][z][0] = 0.0;
      phagecube[x][y][z][1] = 0.0;
      phagecube[x][y][z][2] = 0.0;
      return;
   }

   dim = dim / 3;

   for(i = 0; i < 3; i++)
      for( j = 0; j < 3; j++)
         for(k = 0; k < 3; k++){
            if(i != 1 && j != 1) continue;
            if(i != 1 && k != 1) continue;
            if(k != 1 && j != 1) continue;
            generateFractal(x + i * dim, y + j * dim, z + k * dim, dim);
         }
}

/* Create screen size rectangulare square */
void createSquare(){
   GLfloat v[] = {-1.0, -1.0, 1.0, -1.0, 1.0, 1.0, -1.0, 1.0};

   glGenVertexArrays(1, &square);
   glBindVertexArray(square);
   glGenBuffers(1, &squarevbo);
   glBindBuffer(GL_ARRAY_BUFFER, squarevbo);
   glBufferData(GL_ARRAY_BUFFER, 8 * sizeof(GLfloat), v, GL_STATIC_DRAW);

   glBindVertexArray(0);
}

void InitBiopunk(int width, int height){
   FILE *vbofile;
   FILE *scenefile;
   char line[LINESIZE];
   unsigned int i,j = 0;
   char filename[] = "data/?.obj";

   w = width;
   h = height;
   /* OpenGL related initialization */
   glClearColor(0.0, 0.0, 0.0, 0.0);
   glEnable(GL_DEPTH_TEST);
   perspectiveMatrix(perspectivematrix, 35.0, (float)width/(float)height, 0.1, 1500.0);
   
   /* Load meshes */
   kitchen    = loadOBJ("data/kitchen2.obj");
   walls      = loadOBJ("data/kitchenwall.obj");
   rat        = loadOBJ("data/rat.obj");
   dnasegment = loadOBJ("data/weight.obj");
   tube       = loadOBJ("data/tube.obj");
   phage      = loadOBJ("data/bacteriophage.obj");
   ecoli      = loadOBJ("data/coli2.obj");
   logo1      = loadOBJ("data/logo1.obj");
   logo2      = loadOBJ("data/logo2.obj");

   /* Load letters */
   for(i = 0; i < LETTERS; i++){
      filename[5] = 97+i;
      letters[i] = loadOBJ(filename);
   }

   /* Load vertex array */
   sniff = malloc(sizeof(GLfloat) * rat->vcount);
   vbofile = fopen("data/ratvertex.txt", "r");
   for(i = 0; i < rat->vcount; i++){
      fgets(line, LINESIZE, vbofile);
      sniff[i] = atof(line);
   }
   fclose(vbofile);

   /* Load E. coli scene file */
   scenefile = fopen("data/coliscene.txt", "r");
   for(i = 0; i < 127; i++){
      fgets(line, LINESIZE, scenefile);
      coliscene[j++] = atof(strtok(line, " "));
      coliscene[j++] = atof(strtok(NULL, " "));
      coliscene[j++] = atof(strtok(NULL, " "));
      coliscene[j++] = atof(strtok(NULL, " "));
      coliscene[j++] = atof(strtok(NULL, " "));
      coliscene[j++] = atof(strtok(NULL, " \n"));
   }
   fclose(scenefile);

   /* Generate phage scene */
   for(i = 0; i < PHAGENUM; i++){
      phagescene[i*6]     = drand48() * 10.0 - 5.0;
      phagescene[i*6 + 1] = drand48() * 8.0 - 1.0;
      phagescene[i*6 + 2] = drand48() * 3.0 + 1.34;
      phagescene[i*6 + 3] = drand48() * 10.0 - 5.0;
      phagescene[i*6 + 4] = drand48() * 8.0 - 1.0;
      phagescene[i*6 + 5] = drand48() * 3.0 + 1.34;
   }

   /* Set speed for phages in the phagecube */
   randomSpeed();
   /* Interesting phages get zero to not move */
   generateFractal(0, 0, 0, PHAGECUBESIZE);

   /* Load DNA molecula */
   loadPDB(&dna, "data/2IEF.pdb");
   generateMesh(&dna);
   /* Create screen size square for fragment shaders */
   createSquare();
   /* Init dividing cells */
   initCells();

   /* Bone animation */
   bone = createBone(0,0,0,0,0,0);
   loadVertexWeight(bone, "data/ratbone.txt");
   fillBoneVertex(bone, rat->vertices);

   /* Copy to graphics card */
   glGenBuffers(1, &sniffvbo);
   glBindBuffer(GL_ARRAY_BUFFER, sniffvbo);
   glBufferData(GL_ARRAY_BUFFER, sizeof(GLfloat) * rat->vcount, sniff, GL_STATIC_DRAW);

   /* Load shaders */
   kitchen_src[0] = loadShader(GL_VERTEX_SHADER, "shaders/kitchen.vs");
   kitchen_src[1] = loadShader(GL_FRAGMENT_SHADER, "shaders/kitchen.fs");
   kitchen_shader = createProgram(2, kitchen_src);
   glBindVertexArray(kitchen->vao);
   bindVarToBuff(kitchen_shader, "vertex", kitchen->vbo[MESHVERTEXINDEX], 3);
   bindVarToBuff(kitchen_shader, "normal", kitchen->vbo[MESHNORMALINDEX], 3);
   bindVarToBuff(kitchen_shader, "uvcord", kitchen->vbo[MESHTEXINDEX],    2);

   glBindVertexArray(walls->vao);
   bindVarToBuff(kitchen_shader, "vertex", walls->vbo[MESHVERTEXINDEX], 3);
   bindVarToBuff(kitchen_shader, "normal", walls->vbo[MESHNORMALINDEX], 3);
   bindVarToBuff(kitchen_shader, "uvcord", walls->vbo[MESHTEXINDEX],    2);

   rat_src[0] = loadShader(GL_VERTEX_SHADER, "shaders/rat.vs");
   rat_src[1] = loadShader(GL_FRAGMENT_SHADER, "shaders/rat.fs");
   rat_shader = createProgram(2, rat_src);
   glBindVertexArray(rat->vao);
   bindVarToBuff(rat_shader, "vertex", rat->vbo[MESHVERTEXINDEX], 3);
   bindVarToBuff(rat_shader, "normal", rat->vbo[MESHNORMALINDEX], 3);
   bindVarToBuff(rat_shader, "uvcord", rat->vbo[MESHTEXINDEX],    2);
   bindVarToBuff(rat_shader, "sniff", sniffvbo, 1);

   dnatwist_src[0] = loadShader(GL_VERTEX_SHADER, "shaders/dnatwist.vs");
   dnatwist_src[1] = loadShader(GL_FRAGMENT_SHADER, "shaders/dnatwist.fs");
   dnatwist_shader = createProgram(2, dnatwist_src);
   glBindVertexArray(dnasegment->vao);
   bindVarToBuff(dnatwist_shader, "vertex", dnasegment->vbo[MESHVERTEXINDEX], 3);
   bindVarToBuff(dnatwist_shader, "normal", dnasegment->vbo[MESHNORMALINDEX], 3);

   glBindVertexArray(tube->vao);
   bindVarToBuff(dnatwist_shader, "vertex", tube->vbo[MESHVERTEXINDEX], 3);
   bindVarToBuff(dnatwist_shader, "normal", tube->vbo[MESHNORMALINDEX], 3);

   glBindVertexArray(ecoli->vao);
   bindVarToBuff(dnatwist_shader, "vertex", ecoli->vbo[MESHVERTEXINDEX], 3);
   bindVarToBuff(dnatwist_shader, "normal", ecoli->vbo[MESHNORMALINDEX], 3);

   glBindVertexArray(phage->vao);
   bindVarToBuff(dnatwist_shader, "vertex", phage->vbo[MESHVERTEXINDEX], 3);
   bindVarToBuff(dnatwist_shader, "normal", phage->vbo[MESHNORMALINDEX], 3);

   glBindVertexArray(dna.vao);
   bindVarToBuff(dnatwist_shader, "vertex", dna.vbo[0], 3);
   bindVarToBuff(dnatwist_shader, "normal", dna.vbo[1], 3);

   letter_src[0] = loadShader(GL_VERTEX_SHADER, "shaders/letter.vs");
   letter_src[1] = loadShader(GL_FRAGMENT_SHADER, "shaders/letter.fs");
   letter_shader = createProgram(2, letter_src);
   for(i = 0; i < LETTERS; i++){
      glBindVertexArray(letters[i]->vao);
      bindVarToBuff(letter_shader, "vertex", letters[i]->vbo[MESHVERTEXINDEX], 3);
      bindVarToBuff(letter_shader, "normal", letters[i]->vbo[MESHNORMALINDEX], 3);
      bindVarToBuff(letter_shader, "tcoord", letters[i]->vbo[MESHTEXINDEX],    2);
   }

   glBindVertexArray(logo1->vao);
   bindVarToBuff(dnatwist_shader, "vertex", logo1->vbo[MESHVERTEXINDEX], 3);
   bindVarToBuff(dnatwist_shader, "normal", logo1->vbo[MESHNORMALINDEX], 3);

   glBindVertexArray(logo2->vao);
   bindVarToBuff(dnatwist_shader, "vertex", logo2->vbo[MESHVERTEXINDEX], 3);
   bindVarToBuff(dnatwist_shader, "normal", logo2->vbo[MESHNORMALINDEX], 3);

   square_src[0] = loadShader(GL_VERTEX_SHADER, "shaders/square.vs");
   square_src[1] = loadShader(GL_FRAGMENT_SHADER, "shaders/square.fs");
   square_shader = createProgram(2, square_src);
   glBindVertexArray(square);
   bindVarToBuff(square_shader, "vertex", squarevbo, 2);

   twist_src[0] = square_src[0];
   twist_src[1] = loadShader(GL_FRAGMENT_SHADER, "shaders/square2.fs");
   twist_shader = createProgram(2, twist_src);
   glBindVertexArray(square);
   bindVarToBuff(twist_shader, "vertex", squarevbo, 2);

   dof_src[0] = loadShader(GL_VERTEX_SHADER, "shaders/dof.vs");
   dof_src[1] = loadShader(GL_FRAGMENT_SHADER, "shaders/dof.fs");
   dof_shader = createProgram(2, dof_src);
   glBindVertexArray(square);
   bindVarToBuff(dof_shader, "vertex", squarevbo, 2);

   cell_src[0] = square_src[0];
   cell_src[1] = loadShader(GL_FRAGMENT_SHADER, "shaders/cell.fs");
   cell_shader = createProgram(2, cell_src);
   glBindVertexArray(square);
   bindVarToBuff(cell_shader, "vertex", squarevbo, 2);

   gol_src[0] = square_src[0];
   gol_src[1] = loadShader(GL_FRAGMENT_SHADER, "shaders/gol.fs");
   gol_shader = createProgram(2, gol_src);
   glBindVertexArray(square);
   bindVarToBuff(gol_shader, "vertex", squarevbo, 2);

   simple_src[0] = square_src[0];
   simple_src[1] = loadShader(GL_FRAGMENT_SHADER, "shaders/simple.fs");
   simple_shader = createProgram(2, simple_src);
   glBindVertexArray(square);
   bindVarToBuff(simple_shader, "vertex", squarevbo, 2);

   /* Framebuffer */
   glGenFramebuffers(1, &framebuffer);
   glBindFramebuffer(GL_FRAMEBUFFER, framebuffer);
   glGenTextures(1, &depthbuffer);
   glBindTexture(GL_TEXTURE_2D, depthbuffer);
   glTexImage2D(GL_TEXTURE_2D, 0, GL_DEPTH_COMPONENT24, width, height, 0, GL_DEPTH_COMPONENT, GL_UNSIGNED_BYTE, NULL);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
   glFramebufferTexture2D(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, GL_TEXTURE_2D, depthbuffer, 0);
   glGenTextures(1, &rendertexture);
   glBindTexture(GL_TEXTURE_2D, rendertexture);
   glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, NULL);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
   glGenerateMipmap(GL_TEXTURE_2D);
   glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, rendertexture, 0);
   glBindFramebuffer(GL_FRAMEBUFFER, 0);

   /* Textures */
   glGenTextures(1, &goltex);/* Game of life texture */
   glBindTexture(GL_TEXTURE_2D, goltex);
   glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, NULL);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);

   kitchentex = loadPNGTexture("data/kitchen.png");
   glBindTexture(GL_TEXTURE_2D, kitchentex);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);

   walltex    = loadPNGTexture("data/wall.png");
   glBindTexture(GL_TEXTURE_2D, walltex);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);

   rattex     = loadPNGTexture("data/rat.png");
   glBindTexture(GL_TEXTURE_2D, rattex);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);

   still      = loadPNGTexture("data/still.png");
   glBindTexture(GL_TEXTURE_2D, still);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);

   simplemessage = loadPNGTexture("data/simplemsg.png");
   glBindTexture(GL_TEXTURE_2D, simplemessage);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);

   checkmessage = loadPNGTexture("data/checkmsg.png");
   glBindTexture(GL_TEXTURE_2D, simplemessage);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);

   /* Showtime */
   startTime();
}

void FreeBiopunk(){
   int i;

   freeMesh3D(kitchen);
   freeMesh3D(walls);
   freeMesh3D(rat);
   freeMesh3D(dnasegment);
   freeMesh3D(tube);
   freeMesh3D(ecoli);
   freeMesh3D(phage);
   freePDB(&dna);
   freeMesh3D(logo1);
   freeMesh3D(logo2);
   for(i = 0; i < LETTERS; i++){
      freeMesh3D(letters[i]);
   }
   deleteProgram(kitchen_shader, 2, kitchen_src);
   deleteProgram(rat_shader, 2, rat_src);
   deleteProgram(dnatwist_shader, 2, dnatwist_src);
   deleteProgram(letter_shader, 2, letter_src);
   deleteProgram(square_shader, 2, square_src);
   deleteProgram(twist_shader, 2, twist_src);
   deleteProgram(dof_shader, 2, dof_src);
   deleteProgram(cell_shader, 2, cell_src);
   glDeleteFramebuffers(1, &framebuffer);
   glDeleteRenderbuffers(1, &depthbuffer);
   deleteCells();
   deleteBone(bone);
}

void drawKitchen(float cameramatrix[16], int mode){
   float rotmatrix[16];
   float locmatrix[16];
   float tmpmatrix[16];
   float modelmatrix[16];
   GLuint uniloc;

   rotate(rotmatrix, 0.0, 0.0, 0.0);   
   translate(locmatrix, 0.0, 0.0, 0.0);
   matrixMultiply4x4(rotmatrix, locmatrix, tmpmatrix);
   matrixMultiply4x4(tmpmatrix, cameramatrix, modelmatrix);
   glActiveTexture(GL_TEXTURE0);
   glBindTexture(GL_TEXTURE_2D, walltex);
   glUseProgram(kitchen_shader);
   uniloc = glGetUniformLocation(kitchen_shader, "mvpmatrix");
   glUniformMatrix4fv(uniloc, 1, GL_FALSE, perspectivematrix);
   uniloc = glGetUniformLocation(kitchen_shader, "cammatrix");
   glUniformMatrix4fv(uniloc, 1, GL_FALSE, modelmatrix);
   uniloc = glGetUniformLocation(kitchen_shader, "image");
   glUniform1i(uniloc, 0);
   uniloc = glGetUniformLocation(kitchen_shader, "mode");
   glUniform1i(uniloc, mode);

   drawMesh3D(walls);

   glActiveTexture(GL_TEXTURE0);
   glBindTexture(GL_TEXTURE_2D, kitchentex);
   drawMesh3D(kitchen);
}

void drawRat(float cameramatrix[16], double time, int mode){
   float locmatrix[16];
   float rotmatrix[16];
   float tmpmatrix[16];
   float modelmatrix[16];
   GLuint uniloc;

   translate(locmatrix, -0.179941, 2.483524, -0.6337392);
   rotate(rotmatrix, -0.033239, -3.2438 + M_PI, -0.236697);
   matrixMultiply4x4(rotmatrix, locmatrix, tmpmatrix);
   matrixMultiply4x4(tmpmatrix, cameramatrix, modelmatrix);

   glActiveTexture(GL_TEXTURE0);
   glBindTexture(GL_TEXTURE_2D, rattex);
   glUseProgram(rat_shader);
   uniloc = glGetUniformLocation(rat_shader, "pmatrix");
   glUniformMatrix4fv(uniloc, 1, GL_FALSE, perspectivematrix);
   uniloc = glGetUniformLocation(rat_shader, "modelmatrix");
   glUniformMatrix4fv(uniloc, 1, GL_FALSE, modelmatrix);
   uniloc = glGetUniformLocation(rat_shader, "time");
   glUniform1f(uniloc, time);
   uniloc = glGetUniformLocation(rat_shader, "image");
   glUniform1i(uniloc, 0);
   uniloc = glGetUniformLocation(rat_shader, "mode");
   glUniform1i(uniloc, mode);

   drawMesh3D(rat); 
}

void kitchen_stage1(double time){
   float cameramatrix[16];

   float eye[]    = {1.21821, -0.283882, 1.185586};
   float target[] = {0.0, 0.0, 0.0};
   float up[]     = {0.0, 0.0, 1.0};

   target[1] = logf(time + 1.0) / 2.0;
   lookAt(cameramatrix, eye, target, up);
   drawKitchen(cameramatrix, NORMALMODE);
   drawRat(cameramatrix, time, NORMALRAT);
}

void kitchen_stage2(double time){
   float cameramatrix[16];

   float eye[]    = {0.079734, 1.65343, -0.06565};
   float target[] = {-0.793, 0.891, -0.533};
   float up[]     = {0.0, 0.0, 1.0};

   target[1] = 0.891 + logf(time + 1.0) / 2.0;
   lookAt(cameramatrix, eye, target, up);
   drawKitchen(cameramatrix, NORMALMODE);
   drawRat(cameramatrix, time, NORMALRAT);
}

void kitchen_ratview(double time){
   float cameramatrix[16];

   float eye[]    = {-0.16117, 2.77948, -0.60509};
   float target[] = { -0.179941, 2.483524, -0.6337392};
   float up[]     = {0.0, 0.0, 1.0};
   float rotmatrix[16];
   GLfloat *ptr;

   rotate(rotmatrix, sin(sin(time)*3.14) / 23.0, 0.0, sin(sin(time*2.0)*3.14) / 23.0);
   glBindBuffer(GL_ARRAY_BUFFER, rat->vbo[MESHVERTEXINDEX]);
   ptr = glMapBuffer(GL_ARRAY_BUFFER, GL_READ_WRITE);
   recalculateVertex(bone, ptr, rotmatrix);
   glUnmapBuffer(GL_ARRAY_BUFFER);

   lookAt(cameramatrix, eye, target, up);
   drawKitchen(cameramatrix, NORMALMODE);
   drawRat(cameramatrix, time, NORMALRAT);
}

void kitchen_lampblink(double time){
   float cameramatrix[16];

   float eye[]    = {0.23894, 1.2241, 0.45947};
   float target[] = {0.45368, 0.5262, 1.15525};
   float up[]     = {0.0, 0.0, 1.0};

   lookAt(cameramatrix, eye, target, up);
   if( (time > 0.560 && time < 0.660 ) || 
       (time > 1.650 && time < 1.750) ||
       (time > 2.49 && time < 2.59) ||
       (time > 3.19 && time < 3.29)){
      drawKitchen(cameramatrix, GREENLIGHT);
   }
   else{
      drawKitchen(cameramatrix, NORMALMODE);
   }
}

void kitchen_lampblink2(double time){
   float cameramatrix[16];

   float eye[]    = {0.5586, 2.53683, 1.24308};
   float target[] = { -0.490677, 2.13949, -0.637675};
   float up[]     = {0.0, 0.0, 1.0};

   target[1] = 2.13949 - time / 2.0;
   lookAt(cameramatrix, eye, target, up);

   if( (time > 0.355 && time < 0.455 ) || 
       (time > 1.425 && time < 1.525) ||
       (time > 2.295 && time < 2.395) ||
       (time > 3.135 && time < 3.235)){
      drawKitchen(cameramatrix, GREENLIGHT);
      drawRat(cameramatrix, time, GREENRAT);
   }
   else{
      drawKitchen(cameramatrix, NORMALMODE);
      drawRat(cameramatrix, time, NORMALRAT);
   }
}

/*Hey, it is not fisheye! */
void kitchen_fisheye(double time){
   float cameramatrix[16];
   float eye[]    = {1.0097, -1.6825, -1.350};
   float target[] = {-0.179, 2.393, -1.1};
   float up[]     = {0.0, 0.0, 1.0};
   float pos[3];
   int i;
   float percent;

   percent = time / 8.940;

   for(i = 0; i < 2; i++){
      pos[i] = eye[i] + (target[i] - eye[i]) * percent;
   }
   pos[2] = eye[2];

   lookAt(cameramatrix, pos, target, up);
   drawKitchen(cameramatrix, BUGPERSPECTIVE);
}

void kitchen_in_green1(double time){
   float cameramatrix[16];
   float eye[] = {0.403, 1.077, 0.526};
   float target[] = {-0.694, 2.180, -1.139};
   float up[] = {0.0, 0.0, 1.0};
   float percent;

   percent = time / 9.948;
   eye[1] = eye[1] + percent * (2.18 - 1.077);

   lookAt(cameramatrix, eye, target, up);
   drawKitchen(cameramatrix, GREENLIGHT);
   drawRat(cameramatrix, time, GREENRAT);
}

void kitchen_in_green2(double time){
   float cameramatrix[16];
   float eye[] = {0.403, 1.077, 0.526};
   float target[] = {-0.694, 2.180, -0.32};
   float up[] = {0.0, 0.0, 1.0};
   float percent;

   percent = 1.0 - (time / 9.948);
   eye[1] = eye[1] + percent * (2.18 - 1.077);
   eye[2] = eye[2] + percent * (-0.32 - 0.526);

   lookAt(cameramatrix, eye, target, up);
   drawKitchen(cameramatrix, GREENLIGHT);
   drawRat(cameramatrix, time, GREENRAT);
}

double factorial(int n){
   int c;
   double result = 1.0;

   for(c = 1; c <= n; c++) result *= c;
   return(result);
}

double nCr(int L, int k){
   return( factorial(L) / (factorial(k) * factorial(L - k)) );
}

double Blk(int L, int k, float t){
   return( nCr(L, k) * pow(1.0 - t, L - k) * pow(t, k));
}

void go_to_microscope(double time){
float controlpoints[] = {1.037, 1.678, -0.699,
                         0.456, 1.678, -0.694,
                         0.213, 1.664, -0.267,
                        -0.240, 1.662,  0.097,
                        -0.707, 1.652, -0.350};
   float eye[]           = {0.0, 0.0, 0.0};
   float target[]        = {0.0, 0.0, 0.0};
   float up[]            = {0.0, 0.0, 1.0};
   float percent         = time / 3.461;
   int k,j, L = 4;
   float cameramatrix[16];

   if(percent > 1.0) percent = 1.0;

   for(k = 0; k <= L; k++){
      for(j = 0; j < 3; j++){
         eye[j]    += controlpoints[k*3 + j] * Blk(L, k, percent);
         target[j] += controlpoints[k*3 + j] * Blk(L, k, percent + 0.1);
      }
   }

   lookAt(cameramatrix, eye, target, up);
   drawKitchen(cameramatrix, NORMALMODE);
}

/* Draw a lots of E. coli */
void drawColi(float cameramatrix[16]){
   int i;
   GLuint uniloc;
   float rotmatrix[16];
   float tmpmatrix[16];
   float locmatrix[16];
   float modelmatrix[16];

   glUseProgram(dnatwist_shader);
   uniloc = glGetUniformLocation(dnatwist_shader, "perspective");
   glUniformMatrix4fv(uniloc, 1, GL_FALSE, perspectivematrix);
   uniloc = glGetUniformLocation(dnatwist_shader, "modelmatrix");
   for(i = 0; i < 127; i++){
      translate(locmatrix, coliscene[i*6], coliscene[i*6+1], coliscene[i*6+2]);
      rotate(rotmatrix, coliscene[i*6+3], coliscene[i*6+4], coliscene[i*6+5]);
      matrixMultiply4x4(rotmatrix, locmatrix, tmpmatrix);
      matrixMultiply4x4(tmpmatrix, cameramatrix, modelmatrix);
      glUniformMatrix4fv(uniloc, 1, GL_FALSE, modelmatrix);
      drawMesh3D(ecoli);
   }
}

void drawPhage(float cameramatrix[16], float posx, float posy, float posz, float rotx, float roty, float rotz){
   float locmatrix[16];
   float rotmatrix[16];
   float tmpmatrix[16];
   float modelmatrix[16];
   GLuint uniloc;

   translate(locmatrix, posx, posy, posz);
   rotate(rotmatrix, rotx, roty, rotz);
   matrixMultiply4x4(rotmatrix, locmatrix, tmpmatrix);
   matrixMultiply4x4(tmpmatrix, cameramatrix, modelmatrix);
   uniloc = glGetUniformLocation(dnatwist_shader, "modelmatrix");
   glUniformMatrix4fv(uniloc, 1, GL_FALSE, modelmatrix);
   drawMesh3D(phage);
}

void came_from_microscope(double time){
   float controlpoints[] = { 0.673, -13.799, 6.055,
                             0.707, -10.239, 4.009,
                             0.710,  -7.943, 2.616,
                             0.665,  -6.492, 1.808,
                             0.652,  -4.471, 1.578,
                             0.652,  -2.053, 1.650};
   int j,k, L = 5;
   float locmatrix[16];
   float cameramatrix[16];
   float modelmatrix[16];
   GLuint uniloc;
   float percent = time / 3.544;
   float eye[]           = {0.0, 0.0, 0.0};
   float target[]        = {0.0, 0.0, 0.0};
   float up[]            = {0.0, 0.0, 1.0};

   for(k = 0; k <= L; k++){
      for(j = 0; j < 3; j++){
         eye[j]    += controlpoints[k*3 + j] * Blk(L, k, percent);
         target[j] += controlpoints[k*3 + j] * Blk(L, k, percent + 0.1);
      }
   }
   lookAt(cameramatrix, eye, target, up);
   drawColi(cameramatrix);
   /* Draw tube */
   uniloc = glGetUniformLocation(dnatwist_shader, "modelmatrix");
   translate(locmatrix, 0.0, 0.0, 0.0);
   matrixMultiply4x4(locmatrix, cameramatrix, modelmatrix);
   glUniformMatrix4fv(uniloc, 1, GL_FALSE, modelmatrix);
   drawMesh3D(tube);
   /* First appareance of the phage */
   drawPhage(cameramatrix, 1.3246, 5.9736, 2.6697, 1.5525, 0.0, 0.0);
}

void flying_phage_0(double time){
   float eye[]    = {0.652,  -2.053491, 1.649985};
   float target[] = {0.653058, -0.807370, 1.713267};
   float ppos[]   = { 1.3246, 5.9736, 2.6697};
   float up[]     = {0.0, 0.0, 1.0};
   float cameramatrix[16];
   int i;
   float percent = time / 0.477;

   for(i = 0; i < 3; i++){
      target[i] = target[i] + (ppos[i] - target[i]) * percent;
   }

   lookAt(cameramatrix, eye, target, up);
   drawColi(cameramatrix);
   drawPhage(cameramatrix, ppos[0], ppos[1], ppos[2], 1.5525, 0.0, 0.0);
}

void flying_phage_1(double time){
   float eye[]    = {0.652,  -2.053491, 1.649985};
   float target[] = { 1.3246, 5.9736, 2.6697};
   float up[]     = {0.0, 0.0, 1.0};
   float cameramatrix[16];

   target[1] -= time * 3.585;

   lookAt(cameramatrix, eye, target, up);
   drawColi(cameramatrix);
   drawPhage(cameramatrix, target[0], target[1], target[2], 1.5525, time / 10.0, 0.0);
}

void flying_phage_2(double time){
   float eye[]      = {0.5810, 4.1344, 7.1969};
   float phagepos[] = { 0.9805, 5.3655, 2.8517};
   float target[]   = {0.5810, 4.1344, 2.8517};
   float up[]       = {1.0, 0.0, 0.0};
   float cameramatrix[16];

   target[1]   -= time * 2.585;
   phagepos[1] -= time * 3.585;
   eye[1]       = target[1];

   lookAt(cameramatrix, eye, target, up);
   drawColi(cameramatrix);
   drawPhage(cameramatrix, phagepos[0], phagepos[1], phagepos[2], 1.5525, time / 10.0, 0.0);
}

void lots_of_phage(double time){
   static float eye[] = {8.4037, 10.10545, 8.81592};
   float target[] = {1.49, 3.37, -0.53};
   float up[] = {0.0, 0.0, 1.0};
   float cameramatrix[16];
   int i;
   float percent = time / 10.0;
   float posx, posy, posz, rotx, roty, rotz;
   float dx, dy, dz;
   static int shakeit = 0;

   /* Random camera movements */
   if( ((int)time % 2 == 0) && (shakeit == 0) ){
      eye[0] = drand48() * 6.0 + 2.0;
      eye[1] = drand48() * 6.0 + 4.0;
      shakeit = 1;
   }
   else if( (int)time % 2 != 0){
      shakeit = 0;
   }

   lookAt(cameramatrix, eye, target, up);
   drawColi(cameramatrix);

   for(i = 0; i < PHAGENUM; i++){

      dx = phagescene[i*6+3] - phagescene[i*6];
      dy = phagescene[i*6+4] - phagescene[i*6 + 1];
      dz = phagescene[i*6+5] - phagescene[i*6 + 2];

      posx = phagescene[i*6]     + dx * percent;
      posy = phagescene[i*6 + 1] + dy * percent;
      posz = phagescene[i*6 + 2] + dz * percent;
      rotx = atan(dy / dz);
      roty = atan(dz / dx);
      rotz = atan(dy / dx);

      drawPhage(cameramatrix,posx, posy, posz, rotx, roty, rotz);
   }
}

void lissajou_phage(double time){
   float eye[]    = {3.9742, -10.4917, 3.859};
   float target[] = {2.94, -4.57, 2.40};
   float up[]     = {0.0, 0.0, 1.0};
   float cameramatrix[16];
   float phase;
   float posx, posy, posz;

   eye[0] += time * 0.7;

   lookAt(cameramatrix, eye, target, up);
   drawColi(cameramatrix);

   for(phase = M_PI + time / 5.0; phase < M_PI * 20.0 + time / 5.0; phase += (M_PI * 20.0) / 1300.0){
      posx = sin(phase / 0.2) * 1.5 + target[0];
      posy = sin(phase / 0.3) * 1.5 + target[1];
      posz = sin(phase / 0.4) * 1.2 + target[2];

      drawPhage(cameramatrix, posx, posy, posz, 0.0, 0.0, 0.0);
   }
}

void torusknot_phage(double time){
   float eye[]    = {-6.9232, -2.7019, 3.859};
   float target[] = {-7.9574, 3.2198, 2.40};
   float up[]     = {0.0, 0.0, 1.0};
   float cameramatrix[16];
   float phase;
   float posx, posy, posz;
   float p = 2.0;
   float q = 3.0;
   float r;

   eye[1] += time * 0.7;

   lookAt(cameramatrix, eye, target, up);
   drawColi(cameramatrix);

   for(phase = 0.0 + (time / 5.0); phase < (M_PI * 2.0) + (time / 5.0); phase += (M_PI * 2.0) / 130.0){
      r = cos(q * phase) + 2.0;
      posx = cos(phase * p) * r + target[0];
      posy = sin(phase * p) * r + target[1];
      posz = sin(phase * q)     + target[2];

      drawPhage(cameramatrix, posx, posy, posz, 0.0, 0.0, 0.0);
   }

}

void phagefield(double time){
#define BOMBDELAY 2.0
   float eye[] = {-8.9775, -4.1404, 3.859};
   float target[] = {-7.319, -0.207, 2.644};
   float up[]     = {0.0, 0.0, 1.0};
   float cameramatrix[16];
   float x,y,z;
   int i,j,k;

   target[0] = -8.629 + (PHAGECUBESIZE * 0.1) / 2.0;
   target[1] = -1.837 + (PHAGECUBESIZE * 0.1) / 2.0;
   target[2] =  2.134 + (PHAGECUBESIZE * 0.1) / 2.0;

   eye[0] = -7.319 + cos(time / 10.0) * 4.26;
   eye[1] = -0.207 + sin(time / 10.0) * 4.26;

   /* Use different time for the explosion */
   if(time < BOMBDELAY){
      time = 0.0;
   }
   else{
      time = time - BOMBDELAY;
   }

   lookAt(cameramatrix, eye, target, up);
   drawColi(cameramatrix);

   for(i = 0; i < PHAGECUBESIZE; i++)
      for(j = 0; j < PHAGECUBESIZE; j++)
         for(k = 0; k < PHAGECUBESIZE; k++){
            x = -8.629 + (float)i * 0.1 + time * phagecube[i][j][k][0];
            y = -1.837 + (float)j * 0.1 + time * phagecube[i][j][k][1];
            z =  2.134 + (float)k * 0.2 + time * phagecube[i][j][k][2];
            drawPhage(cameramatrix, x, y, z, time * phagecube[i][j][k][0],
                                             time * phagecube[i][j][k][1],
                                             time * phagecube[i][j][k][2]);
         }
}

void drawspiral(GLuint uniloc, float offset, float torsion, float x, float z){
   float locmatrix[16], rotmatrix[16], modelmatrix[16];

   translate(locmatrix, x, -39.0 + offset, z);
   rotate(rotmatrix, 0.0, offset * torsion, 0.0);
   matrixMultiply4x4(rotmatrix, locmatrix, modelmatrix);
   glUniformMatrix4fv(uniloc, 1, GL_FALSE, modelmatrix);
   drawMesh3D(dnasegment);
}

void dnatwist(double time){
   float offset;
   float torsion[4];
   GLuint uniloc;

   torsion[0] = sinf(time)       * 18.0 * M_PI / 180.0;
   torsion[1] = sinf(time * 0.8) * 18.0 * M_PI / 180.0;
   torsion[2] = sinf(time * 1.2) * 18.0 * M_PI / 180.0;
   torsion[3] = sinf(time * 0.5) * 10.0 * M_PI / 180.0;

   /* Render to texture */
   glBindFramebuffer(GL_FRAMEBUFFER, framebuffer);
   glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

   glUseProgram(dnatwist_shader);
   uniloc = glGetUniformLocation(dnatwist_shader, "perspective");
   glUniformMatrix4fv(uniloc, 1, GL_FALSE, perspectivematrix);
   uniloc = glGetUniformLocation(dnatwist_shader, "modelmatrix");
   for(offset = 0.0; offset < 80.0; offset += 1.0){
      drawspiral(uniloc, offset, torsion[0], -10.0, -127.0);
      drawspiral(uniloc, offset, torsion[1], -15.0, -80.0);
      drawspiral(uniloc, offset, torsion[2], -30.0, -107.0);
      drawspiral(uniloc, offset, torsion[3], -35.0, -97.0);
   }

   /* Draw pictures */
   glBindFramebuffer(GL_FRAMEBUFFER, 0);

   glActiveTexture(GL_TEXTURE0);
   glBindTexture(GL_TEXTURE_2D, simplemessage);
   glActiveTexture(GL_TEXTURE1);
   glBindTexture(GL_TEXTURE_2D, rendertexture);

   glUseProgram(twist_shader);
   uniloc = glGetUniformLocation(square_shader, "still");
   glUniform1i(uniloc, 0);
   uniloc = glGetUniformLocation(square_shader, "geometry");
   glUniform1i(uniloc, 1);
   glBindVertexArray(square);
   glBindBuffer(GL_ARRAY_BUFFER, squarevbo);
   glDrawArrays(GL_TRIANGLE_FAN, 0, 4);

}

void realdna(double time){
   float locmatrix[16];
   float rotmatrix[16];
   float modelmatrix[16];
   float cameramatrix[16];
   float eye[]    = {0.0, -120.0, 0.0};
   float target[] = {-1.19, 60.39, 16.26};
   float up[]     = {0.0, 0.0, 1.0};
   GLuint uniloc;

   glBindFramebuffer(GL_FRAMEBUFFER, framebuffer);
   glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

   lookAt(cameramatrix, eye, target, up);
   /*bounding box: -40.698002 18.539000 -1.595000 39.511002 102.978996 31.021000*/
   glUseProgram(dnatwist_shader);
   uniloc = glGetUniformLocation(dnatwist_shader, "perspective");
   glUniformMatrix4fv(uniloc, 1, GL_FALSE, perspectivematrix);
   uniloc = glGetUniformLocation(dnatwist_shader, "modelmatrix");
   translate(locmatrix, -34.5, 0.956, -54.93);
   rotate(rotmatrix, 0.0, 0.0, time / 2.0);
   matrixMultiply4x4(rotmatrix, cameramatrix, modelmatrix);
   glUniformMatrix4fv(uniloc, 1, GL_FALSE, modelmatrix);
   drawPDB(dna);

   /* Draw images */
   glBindFramebuffer(GL_FRAMEBUFFER, 0);

   glActiveTexture(GL_TEXTURE0);
   glBindTexture(GL_TEXTURE_2D, checkmessage);
   glActiveTexture(GL_TEXTURE1);
   glBindTexture(GL_TEXTURE_2D, rendertexture);

   glUseProgram(twist_shader);
   uniloc = glGetUniformLocation(square_shader, "still");
   glUniform1i(uniloc, 0);
   uniloc = glGetUniformLocation(square_shader, "geometry");
   glUniform1i(uniloc, 1);
   glBindVertexArray(square);
   glBindBuffer(GL_ARRAY_BUFFER, squarevbo);
   glDrawArrays(GL_TRIANGLE_FAN, 0, 4);
}

void overthedna(double time){
   float cameramatrix[16];
   GLuint uniloc;
   float controlpoints[] = {49.799, 43.348, 36.318,
                            41.794, 38.457, 26.750,
                            35.524, 34.265, 19.410,
                            21.382, 25.264,  3.961,
                            5.725,  26.115,  3.699,
                           -0.371,  38.465,  3.982,
                           -6.241,  49.248, 16.605,
                            -7.804, 56.120, 10.255,
                           -14.463, 61.936, 8.492,
                           -21.861, 66.022, 7.396,
                           -27.253, 71.338, 12.759,
                           -35.686, 75.703, 20.118,
                           -23.320, 78.554, 27.965};
   float eye[]    = {0.0, 0.0, 0.0};
   float target[] = {0.0, 0.0, 0.0};
   float up[]     = {0.0, 0.0, 1.0};
   int j,k,L = 12;
   float percent = time / 55.095;

   for(k = 0; k <= L; k++){
      for(j = 0; j < 3; j++){
         eye[j]    += controlpoints[k*3 + j] * Blk(L, k, percent) * 1.7;
         target[j] += controlpoints[k*3 + j] * Blk(L, k, percent + 0.1);
      }
   }

   lookAt(cameramatrix, eye, target, up);
   glUseProgram(dnatwist_shader);
   uniloc = glGetUniformLocation(dnatwist_shader, "perspective");
   glUniformMatrix4fv(uniloc, 1, GL_FALSE, perspectivematrix);
   uniloc = glGetUniformLocation(dnatwist_shader, "modelmatrix");
   glUniformMatrix4fv(uniloc, 1, GL_FALSE, cameramatrix);
   drawPDB(dna);

}

void credits(double time){
   GLuint uniloc;
   char code[] = "travis";
   char grfx[] = "grass";
   char musi[] = "shamen";
   float locmatrix[16];
   float rotmatrix[16];
   float modelmatrix[16];
   unsigned int i;
   float dummy;
   float ypos, zpos;

   glBindFramebuffer(GL_FRAMEBUFFER, framebuffer);
   glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

   ypos = sin(time) * 10.0;
   zpos = -70.0 + cos(time) * 10.0;

   glUseProgram(letter_shader);
   uniloc = glGetUniformLocation(letter_shader, "perspective");
   glUniformMatrix4fv(uniloc, 1, GL_FALSE, perspectivematrix);

   uniloc = glGetUniformLocation(letter_shader, "cammatrix");

   for(i = 0; i < strlen(code); i++){
      if(i < 4) dummy = 7.0; else dummy = 6.6;
      translate(locmatrix, -21.0 + i * dummy, ypos, zpos);
      rotate(rotmatrix, 4.712389, 0.0, 0.0 );
      matrixMultiply4x4(rotmatrix, locmatrix, modelmatrix);
      glUniformMatrix4fv(uniloc, 1, GL_FALSE, modelmatrix);
      drawMesh3D(letters[code[i]-97]);
   }

   ypos = sin(time + (M_PI * 2.0 / 3.0)) * 10.0;
   zpos = -70.0 + cos(time + (M_PI * 2.0 / 3.0)) * 10.0;

   for(i = 0; i < strlen(grfx); i++){
      translate(locmatrix, -21.0 + i * 7.0, ypos, zpos);
      rotate(rotmatrix, 4.712389, 0.0, 0.0 );
      matrixMultiply4x4(rotmatrix, locmatrix, modelmatrix);
      glUniformMatrix4fv(uniloc, 1, GL_FALSE, modelmatrix);
      drawMesh3D(letters[grfx[i]-97]);
   }

   ypos = sin(time + (M_PI / 3.0)) * 10.0;
   zpos = -70.0 + cos(time + (M_PI / 3.0)) * 10.0;

   for(i = 0; i < strlen(musi); i++){
      translate(locmatrix, -21.0 + i * 7.0, ypos, zpos);
      rotate(rotmatrix, 4.712389, 0.0, 0.0 );
      matrixMultiply4x4(rotmatrix, locmatrix, modelmatrix);
      glUniformMatrix4fv(uniloc, 1, GL_FALSE, modelmatrix);
      drawMesh3D(letters[musi[i]-97]);
   }

   glBindFramebuffer(GL_FRAMEBUFFER, 0);

   glActiveTexture(GL_TEXTURE0);
   glBindTexture(GL_TEXTURE_2D, still);
   glActiveTexture(GL_TEXTURE1);
   glBindTexture(GL_TEXTURE_2D, rendertexture);

   glUseProgram(square_shader);
   uniloc = glGetUniformLocation(square_shader, "still");
   glUniform1i(uniloc, 0);
   uniloc = glGetUniformLocation(square_shader, "geometry");
   glUniform1i(uniloc, 1);
   glBindVertexArray(square);
   glBindBuffer(GL_ARRAY_BUFFER, squarevbo);
   glDrawArrays(GL_TRIANGLE_FAN, 0, 4);

}

void greetings(double time){
   char text[] = "booze design conspiracy singular crew archee united force vovoid the bad sectors fairlight asd";
   unsigned int i,j,k;
   GLuint uniloc;
   float locmatrix[16];
   float rotmatrix[16];
   float modelmatrix[16];
   float pos[3], target[3];
   float percent;
   float controlpoints[] = {-29.221 ,  20.802, -56.778,
                            -20.103 ,  18.667, -56.778,
                            -13.579 ,  17.656, -56.778,
                             -3.503 ,  13.385, -56.778,
                              2.952 ,   9.416, -56.778,
                              11.962,  14.846, -56.778,
                              24.538,   8.568, -56.778,
                              24.538,   0.092, -46.222,
                              24.538,   2.491, -38.705,
                              10.591,  -0.964, -38.483,
                               8.735,  -1.917, -28.653,
                               3.005,  -5.969, -21.223,
                              -5.590,  -5.302, -28.653,
                             -18.205,  -0.820, -36.063,
                             -10.502,   7.623, -51.004,
                               6.608,   0.495, -51.801,
                              16.284,  -5.585, -51.843,
                              31.486, -14.754, -51.843};
#define CONTROLPOINTS 17

   glBindFramebuffer(GL_FRAMEBUFFER, framebuffer);
   glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

   glUseProgram(letter_shader);
   uniloc = glGetUniformLocation(letter_shader, "perspective");
   glUniformMatrix4fv(uniloc, 1, GL_FALSE, perspectivematrix);

   uniloc = glGetUniformLocation(letter_shader, "cammatrix");

   for(i = 0; i < strlen(text); i++){
      percent = time / 4.0 - (0.1 * (float)i);
      if(percent < 0.0 || percent > 1.0) continue;
      if(text[i] == 32) continue; /* This is space */
      /* Zero variables */
      for(k = 0; k < 3; k++){
         pos[k] = 0.0;
         target[k] = 0.0;
      }

      for(j = 0; j <= CONTROLPOINTS; j++){
         for(k = 0; k < 3; k++){
            pos[k] += controlpoints[j*3+k]   * Blk(CONTROLPOINTS, j, percent);
            target[k] += controlpoints[j*3+k]   * Blk(CONTROLPOINTS, j, percent + 0.1);
         }
      }
      translate(locmatrix, pos[0], pos[1], pos[2]);
      rotate(rotmatrix, 4.712389, 0.0, 0.0);
      matrixMultiply4x4(rotmatrix, locmatrix, modelmatrix);
      glUniformMatrix4fv(uniloc, 1, GL_FALSE, modelmatrix);
      drawMesh3D(letters[text[i]-97]);
   }

   glBindFramebuffer(GL_FRAMEBUFFER, 0);

   glActiveTexture(GL_TEXTURE0);
   glBindTexture(GL_TEXTURE_2D, depthbuffer);
   glActiveTexture(GL_TEXTURE1);
   glBindTexture(GL_TEXTURE_2D, rendertexture);

   glUseProgram(dof_shader);
   uniloc = glGetUniformLocation(dof_shader, "depth");
   glUniform1i(uniloc, 0);
   uniloc = glGetUniformLocation(dof_shader, "geometry");
   glUniform1i(uniloc, 1);
   glBindVertexArray(square);
   glBindBuffer(GL_ARRAY_BUFFER, squarevbo);
   glDrawArrays(GL_TRIANGLE_FAN, 0, 4);
}

void ende(float time){
   float locmatrix[16];
   float rotmatrix[16];
   float cameramatrix[16];
   float modelmatrix[16];
   GLuint uniloc;
   float eye[] = {0.0, 0.0, 0.0};
   float target[] = {0.0, 0.0, -14.266};
   float up[] = {0.0, 1.0, 0.0};
   float yrot;

   yrot = (M_PI / 2.0) + sin(time) * 0.7;

   eye[0] = 0.0     + cos(yrot) * 14.266;
   eye[2] = -14.266 + sin(yrot) * 14.266;

   lookAt(cameramatrix, eye, target, up);

   translate(locmatrix, 0.0, 0.0, -14.266);
   rotate(rotmatrix, 0.0, 0.0, sin(time / 3.0) * 10.0);
   matrixMultiply4x4(rotmatrix, locmatrix, modelmatrix);
   matrixMultiply4x4(modelmatrix, cameramatrix, locmatrix);

   glUseProgram(dnatwist_shader);
   uniloc = glGetUniformLocation(dnatwist_shader, "perspective");
   glUniformMatrix4fv(uniloc, 1, GL_FALSE, perspectivematrix);
   uniloc = glGetUniformLocation(dnatwist_shader, "modelmatrix");
   glUniformMatrix4fv(uniloc, 1, GL_FALSE, locmatrix);
   drawMesh3D(logo1);

   translate(locmatrix, 0.0, 0.0, -14.266);
   rotate(rotmatrix, 0.0, 0.0, -1.0 * sin(time / 3.0) * 10.0);
   matrixMultiply4x4(rotmatrix, locmatrix, modelmatrix);
   matrixMultiply4x4(modelmatrix, cameramatrix, locmatrix);
   glUniformMatrix4fv(uniloc, 1, GL_FALSE, locmatrix);
   drawMesh3D(logo2);
}

void celldivide(double time){
   GLuint uniloc;
   int bc;

   bc = updateCells(time);
   glActiveTexture(GL_TEXTURE0);
   glBindTexture(GL_TEXTURE_2D, celltexture);

   glUseProgram(cell_shader);
   uniloc = glGetUniformLocation(cell_shader, "celldata");
   glUniform1i(uniloc, 0);
   uniloc = glGetUniformLocation(cell_shader, "bornedcellnum");
   glUniform1i(uniloc, bc);

   glBindVertexArray(square);
   glBindBuffer(GL_ARRAY_BUFFER, squarevbo);
   glDrawArrays(GL_TRIANGLE_FAN, 0, 4);
}

/* There is a bug in here so you can't see this part of the demo */
void game_of_life(){
   GLuint uniloc;
   static int mode = 1;

   /* Render to texture */
   glBindFramebuffer(GL_FRAMEBUFFER, framebuffer);
   glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

   glActiveTexture(GL_TEXTURE0);
   glBindTexture(GL_TEXTURE_2D, goltex);

   glUseProgram(gol_shader);
   uniloc = glGetUniformLocation(gol_shader, "source");
   glUniform1i(uniloc, 0);
   uniloc = glGetUniformLocation(gol_shader, "mode");
   glUniform1i(uniloc, mode);

   glBindVertexArray(square);
   glBindBuffer(GL_ARRAY_BUFFER, squarevbo);
   glDrawArrays(GL_TRIANGLE_FAN, 0, 4);

   glBindBuffer(GL_FRAMEBUFFER, 0);

   /* Show rendered texture */
   glActiveTexture(GL_TEXTURE0);
   glBindTexture(GL_TEXTURE_2D, rendertexture);

   glUseProgram( simple_shader );
   uniloc = glGetUniformLocation(simple_shader , "img");
   glUniform1i(uniloc, 0);

   glBindVertexArray(square);
   glBindBuffer(GL_ARRAY_BUFFER, squarevbo);
   glDrawArrays(GL_TRIANGLE_FAN, 0, 4);
 
   /* Copy rendered texture to goftex for the next iteration */
   glReadBuffer(GL_FRONT_LEFT);
   glCopyTexImage2D(goltex, 0, GL_RGBA, 0, 0, w, h, 0);

   if(mode == 1) mode = 0;
}

void PlayBiopunk(){
   double time;

   glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

   time = getTimeInterval();
   if(time >   0.0   && time <   6.927)   kitchen_stage1(time);
   if(time >   6.927 && time <  13.761)   kitchen_stage2(time - 6.927);
   if(time >  13.761 && time <  27.120)   kitchen_ratview(time - 13.761);
   if(time >  27.120 && time <  34.115)   kitchen_lampblink(time - 27.120);
   if(time >  34.115 && time <  41.312)   kitchen_lampblink2(time - 34.115);
   if(time >  41.312 && time <  48.252)   kitchen_fisheye(time - 41.312);
   if(time >  48.252 && time <  55.080)   kitchen_in_green1(time - 48.252);
   if(time >  55.080 && time <  65.028)   kitchen_in_green2(time - 55.080);
   if(time >  65.028 && time <  68.489)   go_to_microscope(time - 65.028);
   if(time >  68.489 && time <  72.033)   came_from_microscope(time - 68.489);
   if(time >  72.033 && time <  72.500)   flying_phage_0(time - 72.033);
   if(time >  72.500 && time <  75.642)   flying_phage_1(time - 72.500);
   if(time >  75.642 && time <  79.099)   flying_phage_2(time - 75.642);
   if(time >  79.099 && time <  89.057)   lots_of_phage(time - 79.099);
   if(time >  89.057 && time <  96.164)   lissajou_phage(time - 89.057);
   if(time >  96.164 && time < 110.039)   torusknot_phage(time - 96.164);
   if(time > 110.039 && time < 123.714)   phagefield(time - 110.039);
   if(time > 123.714 && time < 137.284)   dnatwist(time);
   if(time > 137.284 && time < 150.855)   realdna(time);
   if(time > 150.855 && time < 205.950)   overthedna(time - 150.855);
   if(time > 205.950 && time < 219.640)   credits(time);
   if(time > 219.640 && time < 274.673)   celldivide(time - 219.640);
   if(time > 274.673 && time < 315.915)   greetings(time - 274.673);
   if(time > 315.915)                     ende(time);
}
