#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "matrix.h"
#include "bone.h"

#define LINELEN 100

Bone *createBone(float hx, float hy, float hz, float tx, float ty, float tz){
   Bone *newbone;

   newbone = malloc(sizeof(Bone));

   if(newbone != NULL){
      newbone->headx = hx;
      newbone->heady = hy;
      newbone->headz = hz;

      newbone->tailx = tx;
      newbone->taily = ty;
      newbone->tailz = tz;

      newbone->vertnum = 0;
      newbone->verts = NULL;
   }

   return(newbone);
}

void deleteBone(Bone *b){
   if(b->verts != NULL) free(b->verts);
   free(b);
}

int addVertex(Bone *b, VertexVector v){
   VertexVector *newverts;

   newverts = realloc(b->verts, sizeof(VertexVector) * (b->vertnum+1));
   if(newverts != NULL){
      b->verts = newverts;
      b->verts[b->vertnum] = v;
      b->vertnum++;
      return 1;
   }
   return 0;
}

void loadVertexWeight(Bone *b, const char *filename){
   FILE *file;
   char line[LINELEN];
   char first = 1;
   VertexVector v;

   file = fopen(filename, "r");
   while(fgets(line, LINELEN, file) != NULL){
      if(first){
         strtok(line, " ");
         b->headx = atof(strtok(NULL, " "));
         b->heady = atof(strtok(NULL, " "));
         b->headz = atof(strtok(NULL, " "));
         b->tailx = atof(strtok(NULL, " "));
         b->taily = atof(strtok(NULL, " "));
         b->tailz = atof(strtok(NULL, " \n"));

         first = 0;
      }
      else{
         v.index  = atoi(strtok(line, " "));
         v.weight = atof(strtok(NULL, "\n "));
         addVertex(b, v);
      }
   }
   fclose(file);
}

void fillBoneVertex(Bone *b, float *vbo){
   unsigned int i;
   unsigned int index;

   for(i = 0; i < b->vertnum; i++){
      index = b->verts[i].index;
      /* FIXME wrong Bone tail coordinate export from Blender so quick fix in demo code :-) */
      b->verts[i].dx = vbo[index * 3 + 0]/* - b->tailx*/;
      b->verts[i].dy = vbo[index * 3 + 1] -0.01 ;/* - b->taily*/;
      b->verts[i].dz = vbo[index * 3 + 2]/* - b->tailz*/;
   }
}

void recalculateVertex(Bone *b, float *vbo, float matrix[16]){
   unsigned int i,j;
   float locmatrix[16], resmatrix[16];
   float w;

   translate(locmatrix, b->tailx, b->taily, b->tailz);
   matrixMultiply4x4(matrix, locmatrix, resmatrix);

   for(i = 0; i < b->vertnum; i++){/* Every vertex */
      /* It is a matrix multiply by the position vector
         I have made a simple optimization and the forth component was removed.
      */


      for(j = 0; j < 3; j++){
         vbo[b->verts[i].index * 3 + j] = resmatrix[j*4]     * b->verts[i].dx + 
                                          resmatrix[j*4 + 1] * b->verts[i].dy + 
                                          resmatrix[j*4 + 2] * b->verts[i].dz +
                                          resmatrix[j*4 + 3];
      }
      /* FIXME Well, this weight calculation workaround is for correct the error of the bone weight making procedures */
      w = b->verts[i].weight + 0.9;
      if(w > 1.0) w = 1.0;
      vbo[b->verts[i].index * 3 + 0] = b->verts[i].dx + (vbo[b->verts[i].index * 3 + 0] - b->verts[i].dx) * w;
      vbo[b->verts[i].index * 3 + 1] = b->verts[i].dy + (vbo[b->verts[i].index * 3 + 1] - b->verts[i].dy) * w;
      vbo[b->verts[i].index * 3 + 2] = b->verts[i].dz + (vbo[b->verts[i].index * 3 + 2] - b->verts[i].dz) * w;
   }

}

