#define GL_GLEXT_PROTOTYPES
#include <GL/gl.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include "celldivide.h"

GLuint cellbuffer;
Cells cells;

void initCells(){
   GLubyte *ptr;

   glPixelStorei(GL_UNPACK_ALIGNMENT, 1);

   glGenTextures(1, &celltexture);
   glBindTexture(GL_TEXTURE_2D, celltexture);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
   glTexImage2D(GL_TEXTURE_2D, 0, GL_RED, MAXCELLNUM, CELLATTRIBNUM, 0, GL_RED, GL_UNSIGNED_BYTE, 0);
   glBindTexture(GL_TEXTURE_2D, 0);

   glGenBuffers(1, &cellbuffer);
   glBindBuffer(GL_PIXEL_UNPACK_BUFFER, cellbuffer);
   glBufferData(GL_PIXEL_UNPACK_BUFFER, sizeof(GLubyte) * MAXCELLNUM * CELLATTRIBNUM, 0, GL_DYNAMIC_DRAW);

   /* Init one cell */
   cells.cells = malloc(sizeof(Cell) * MAXCELLNUM);
   cells.bornedcells = 1;
   cells.cells[0].xpos = 0.5;
   cells.cells[0].ypos = 0.5;
   cells.cells[0].xspeed = 0.0;
   cells.cells[0].yspeed = 0.0;
   cells.cells[0].divide = 1.0;
   ptr = glMapBuffer(GL_PIXEL_UNPACK_BUFFER, GL_WRITE_ONLY);
   ptr[0 * MAXCELLNUM] = 127;     /* xpos   */
   ptr[1 * MAXCELLNUM] = 0;       /* xpos   */
   ptr[2 * MAXCELLNUM] = 127;     /* ypos   */
   ptr[3 * MAXCELLNUM] = 0;       /* ypos   */
   glUnmapBuffer(GL_PIXEL_UNPACK_BUFFER);

   glBindBuffer(GL_PIXEL_UNPACK_BUFFER, 0);

}

void deleteCells(){
   glDeleteTextures(1, &celltexture);
   glDeleteBuffers(1, &cellbuffer);
}

/* Need to call it after glBindBuffer */
void updatePixelData(double time){
   GLubyte *ptr;
   unsigned int i;
   unsigned int newcell;
   static double prevtime = 0.0;
   double interval;

   interval = time - prevtime;

   ptr = glMapBuffer(GL_PIXEL_UNPACK_BUFFER, GL_WRITE_ONLY);
   for(i = 0; i < cells.bornedcells; i++){
      cells.cells[i].xpos += cells.cells[i].xspeed * interval;
      cells.cells[i].ypos += cells.cells[i].yspeed * interval;
      cells.cells[i].xspeed += (0.0 - cells.cells[i].xspeed) * 0.001 * interval;
      cells.cells[i].yspeed += (0.0 - cells.cells[i].yspeed) * 0.001 * interval;
      cells.cells[i].divide -= interval;
      if(cells.cells[i].divide < 0.0 && cells.bornedcells < MAXCELLNUM){
         /* Born new cell */
         cells.cells[i].divide = drand48() * 3.0;
         newcell = cells.bornedcells;
         cells.cells[newcell].xpos = cells.cells[i].xpos;
         cells.cells[newcell].ypos = cells.cells[i].ypos;
         cells.cells[newcell].xspeed = drand48() / 10.0 - 0.05;
         cells.cells[newcell].yspeed = drand48() / 10.0 - 0.05;
         cells.cells[newcell].divide = drand48() * 3.0;
         cells.bornedcells++;
      }

      ptr[0 * MAXCELLNUM + i] = (65535.0 * cells.cells[i].xpos) / 255.0;
      ptr[1*  MAXCELLNUM + i] = (int)(65535.0 * cells.cells[i].xpos) % 255;
      ptr[2 * MAXCELLNUM + i] = (65535.0 * cells.cells[i].ypos) / 255.0;
      ptr[3 * MAXCELLNUM + i] = (int)(65535.0 * cells.cells[i].ypos) % 255;
   }
   glUnmapBuffer(GL_PIXEL_UNPACK_BUFFER);

   prevtime = time;
}

int updateCells(double time){
   glBindTexture(GL_TEXTURE_2D, celltexture);
   glBindBuffer(GL_PIXEL_UNPACK_BUFFER, cellbuffer);
   updatePixelData(time);
   /* Copy from PBO to texture */
   glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, MAXCELLNUM, CELLATTRIBNUM, GL_RED, GL_UNSIGNED_BYTE, 0);
   glBindBuffer(GL_PIXEL_UNPACK_BUFFER, 0);
   return(cells.bornedcells);
}


